from __future__ import annotations
from typing import Any, Callable, Generic, List, Optional, TypeVar
from .monad import Monad
from .monoid import Monoid

T = TypeVar("T")
S = TypeVar("S")
E = TypeVar("E")


class Maybe(Monad[T]):
    def __init__(self) -> None:  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def pure(cls, value: T) -> Maybe[T]:
        return Just(value)

    def bind(self, function: Callable[[T], Maybe[S]]) -> Maybe[S]:
        if isinstance(self, Just):
            return function(self.value)
        else:
            new: Maybe[S] = Nothing()
            return new

    def map(self, function: Callable[[T], S]) -> Maybe[S]:
        if isinstance(self, Just):
            return Just(function(self.value))
        else:
            new: Maybe[S] = Nothing()
            return new

    def apply(self, functor: Maybe[Callable[[T], S]]) -> Maybe[S]:
        if isinstance(functor, Just):
            return self.map(functor.value)
        else:
            new: Maybe[S] = Nothing()
            return new

    def withDefault(self, default: T) -> T:
        if isinstance(self, Just):
            return self.value
        else:
            return default

    @classmethod
    def fromResult(cls, m: Result[T, E]) -> Maybe[T]:
        return m.map(Maybe.pure).withDefault(Nothing())

    def toResult(self, error: E) -> Result[T, E]:
        if isinstance(self, Just):
            return Ok(self.value)
        else:
            return Err(error)

    @classmethod
    def fromList(self, xs: List[T]) -> Maybe[T]:
        if xs:
            return Just(xs[0])
        else:
            return Nothing()

    __rshift__ = bind
    __mul__ = __rmul__ = map


class Just(Maybe[T]):
    def __init__(self, value: T) -> None:
        self.value = value

    def __eq__(self, other: object):
        return isinstance(other, Just) and self.value == other.value

    def __repr__(self) -> str:  # pragma: no cover
        return f"<Just {self.value}>"


class Nothing(Maybe[T]):
    def __init__(self) -> None:
        ...

    def __eq__(self, other: object):
        return isinstance(other, Nothing)

    def __repr__(self) -> str:  # pragma: no cover
        return "<Nothing>"


def maybe(value: T, predicate: Optional[Callable[[T], bool]] = None) -> Maybe[T]:
    predicate = predicate or (lambda x: x is not None)
    if predicate(value):
        return Just(value)
    else:
        return Nothing()


class First(Monoid[Maybe[T]]):
    @classmethod
    def mzero(cls) -> First:
        return First(Nothing())

    def mappend(self, other: First):
        if isinstance(self.value, Just):
            return self
        else:
            return other

    def __repr__(self) -> str:  # pragma: no cover
        return f"<First {self.value}>"

    __add__ = mappend


def first(xs: List[Maybe[T]]) -> Maybe[T]:
    return First.mconcat(map(lambda x: First(x), xs)).value


class Last(Monoid[Maybe[T]]):
    @classmethod
    def mzero(cls) -> Last:
        return Last(Nothing())

    def mappend(self, other: Last):
        if isinstance(other.value, Just):
            return other
        else:
            return self

    def __repr__(self) -> str:  # pragma: no cover
        return f"<Last {self.value}>"

    __add__ = mappend


def last(xs: List[Maybe[T]]) -> Maybe[T]:
    return Last.mconcat(map(lambda x: Last(x), xs)).value


# Import Result last to avoid a circular import error
from .result import Result, Ok, Err



from __future__ import annotations
from functools import reduce, update_wrapper
from typing import Any, Callable, Generic, TypeVar

from .monad import Monad

T = TypeVar("T")
S = TypeVar("S")
Env = TypeVar("Env")
F = Callable[[Env], T]


class Reader(Monad[T], Generic[Env, T]):
    def __init__(self, function: F) -> None:
        update_wrapper(self, function)
        self.function = function

    def __call__(self, environment: Env) -> T:
        return self.function(environment)

    @classmethod
    def pure(cls, value: T) -> Reader[Env, T]:
        f: F = lambda x: value
        return cls(f)

    def map(self, function: Callable[[T], S]) -> Reader[Env, S]:
        f: Callable[[Env], S] = lambda x: function(self.function(x))
        return Reader(f)

    def apply(self, r: Reader[Env, Callable[[T], S]]) -> Reader[Env, S]:
        f: Callable[[Env], S] = lambda x: r.function(x)(self(x))
        return Reader(f)

    def bind(self, function: Callable[[T], Reader[Env, S]]) -> Reader[Env, S]:
        f: Callable[[Env], S] = lambda x: function(self.function(x))(x)
        return Reader(f)

    def __eq__(self, other: object):  # pragma: no cover
        return isinstance(other, Reader) and self.function == other.function

    def __repr__(self):  # pragma: no cover
        module = self.function.__module__
        name = self.function.__name__
        return f"<Reader {module}.{name}>"

    __mul__ = __rmul__ = map
    __rshift__ = bind


class Curried(Reader[Env, T]):
    def __call__(self, *args):
        return reduce(lambda f, x: f(x), args, self.function)


def curry(f: Callable):
    def wrapped(args, remaining):
        if remaining == 0:
            return f(*args)
        else:
            curried = lambda x: wrapped(args + [x], remaining - 1)
            return Curried(update_wrapper(curried, f))

    return wrapped([], f.__code__.co_argcount)



from __future__ import annotations
from typing import Any, Callable, Generic, TypeVar

from .monad import Monad

T = TypeVar("T")
S = TypeVar("S")
E = TypeVar("E")


class Result(Monad[T], Generic[T, E]):
    def __init__(self) -> None:  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def pure(cls, value: T) -> Result[T, E]:
        return Ok(value)

    def bind(self, function: Callable[[T], Result[S, E]]) -> Result[S, E]:
        if isinstance(self, Ok):
            return function(self.value)
        elif isinstance(self, Err):
            new: Result[S, E] = Err(self.err)
            return new
        else:  # pragma: no cover
            raise TypeError

    def map(self, function: Callable[[T], S]) -> Result[S, E]:
        if isinstance(self, Ok):
            return Result.pure(function(self.value))
        elif isinstance(self, Err):
            new: Result[S, E] = Err(self.err)
            return new
        else:  # pragma: no cover
            raise TypeError

    def apply(self, functor: Result[Callable[[T], S], E]) -> Result[S, E]:
        if isinstance(functor, Ok):
            return self.map(functor.value)
        elif isinstance(functor, Err):
            new: Result[S, E] = Err(functor.err)
            return new
        else:  # pragma: no cover
            raise TypeError

    def withDefault(self, default: T) -> T:
        if isinstance(self, Ok):
            return self.value
        else:
            return default

    @classmethod
    def fromMaybe(cls, m: Maybe[T], error: E) -> Result[T, E]:
        return m.map(Result.pure).withDefault(Err(error))

    def toMaybe(self) -> Maybe[T]:
        return self.map(Maybe.pure).withDefault(Nothing())

    __rshift__ = bind
    __mul__ = __rmul__ = map


class Ok(Result[T, E]):
    def __init__(self, value: T) -> None:
        self.value = value

    def __eq__(self, other: object):
        return isinstance(other, Ok) and self.value == other.value

    def __repr__(self) -> str:  # pragma: no cover
        return f"<Ok {self.value}>"


class Err(Result[T, E]):
    def __init__(self, err: E) -> None:
        self.err = err

    def __eq__(self, other: object):
        return isinstance(other, Err) and self.err == other.err

    def __repr__(self) -> str:  # pragma: no cover
        return f"<Err {self.err}>"


def safe(function: Callable[..., T]) -> Callable[..., Result[T, Exception]]:
    """Wraps a function that may raise an exception.

    e.g.:
        @safe
        def bad() -> int:
            raise Exception("oops")

    """

    def wrapped(*args, **kwargs) -> Result[T, Exception]:
        try:
            return Ok(function(*args, **kwargs))
        except Exception as e:
            return Err(e)

    return wrapped


# Import Maybe last to avoid a circular import error
from .maybe import Maybe, Nothing



from __future__ import annotations
from typing import Any, Callable, TypeVar

from .functor import Functor

T = TypeVar("T")
S = TypeVar("S")


class Applicative(Functor[T]):
    @classmethod
    def pure(cls, value: T) -> Applicative[T]:  # pragma: no cover
        raise NotImplementedError

    # FIXME: Functor type set to Any, as the proper value
    # (Functor[Callable[[T], S]]) is reported as incompatible with subclass
    # implementations due to a flaw in mypy:
    # https://github.com/python/mypy/issues/1317
    def apply(self, functor: Any) -> Functor[S]:  # pragma: no cover
        raise NotImplementedError



from .functor import Functor
from .applicative import Applicative
from .monad import Monad
from .list import List
from .maybe import Maybe, Just, Nothing
from .result import Result, Ok, Err
from .reader import Reader



from __future__ import annotations
from functools import reduce
from itertools import chain
from typing import Callable, TypeVar

from .monad import Monad
from .monoid import Monoidal

T = TypeVar("T")
S = TypeVar("S")


class List(Monad[T], Monoidal[list]):
    @classmethod
    def pure(cls, value: T) -> List[T]:
        return List([value])

    def bind(self, function: Callable[[T], List[S]]) -> List[S]:
        return reduce(List.mappend, map(function, self.value), List.mzero())

    def map(self, function: Callable[[T], S]) -> List[S]:
        return List(list(map(function, self.value)))

    def apply(self, functor: List[Callable[[T], S]]) -> List[S]:
        return List(
            list(chain.from_iterable([map(f, self.value) for f in functor.value]))
        )

    @classmethod
    def mzero(cls) -> List[T]:
        return cls(list())

    def mappend(self, other: List[T]) -> List[T]:
        return List(self.value + other.value)

    __add__ = mappend
    __mul__ = __rmul__ = map
    __rshift__ = bind



from __future__ import annotations
from typing import Any, Callable, Generic, TypeVar

from .applicative import Applicative
from .functor import Functor

T = TypeVar("T")
S = TypeVar("S")


class Monad(Applicative[T]):
    # FIXME: Callable return type set to Any, as the proper value
    # (Monad[S]) is reported as incompatible with subclass
    # implementations due to a flaw in mypy:
    # https://github.com/python/mypy/issues/1317
    def bind(self, function: Callable[[T], Any]) -> Monad[S]:  # pragma: no cover
        raise NotImplementedError

    __rshift__ = bind



from __future__ import annotations
from functools import reduce
from numbers import Complex
from decimal import Decimal
from typing import Any, Callable, Generic, Iterator, Type, TypeVar, Union

T = TypeVar("T")


class Monoid(Generic[T]):
    def __init__(self, value: T) -> None:
        self.value = value

    # FIXME: Other type set to Any, as the proper value (Monoid[T]) is
    # reported as incompatible with subclass implementations due to a
    # flaw in mypy: https://github.com/python/mypy/issues/1317
    def mappend(self, other: Any) -> Monoid[T]:  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def mzero(cls) -> Monoid[T]:  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def mconcat(cls, xs: Iterator[Monoid[T]]) -> Monoid[T]:
        return reduce(cls.mappend, xs, cls.mzero())

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Monoid)
            and type(self) == type(other)
            and self.value == other.value
        )

    __add__ = mappend


class Monoidal(Monoid[T]):
    def __repr__(self):  # pragma: no cover
        return repr(self.value)


class String(Monoidal[str]):
    @classmethod
    def mzero(cls) -> Monoidal:
        return cls(str())

    def mappend(self, other: String) -> String:
        return String(self.value + other.value)

    __add__ = mappend


class Addition(Monoidal[Union[int, float]]):
    @classmethod
    def mzero(cls) -> Addition:
        return cls(0)

    def mappend(self, other: Addition) -> Addition:
        return Addition(self.value + other.value)

    __add__ = mappend


class Multiplication(Monoidal[Union[int, float]]):
    @classmethod
    def mzero(cls) -> Multiplication:
        return cls(1)

    def mappend(self, other: Multiplication) -> Multiplication:
        return Multiplication(self.value * other.value)

    __add__ = mappend



from setuptools import setup  # type: ignore

with open("README.md", "r") as f:
    long_description = f.read()


setup(
    name="typesafe-monads",
    version="0.4",
    author="Correl Roush",
    author_email="correl@gmail.com",
    description="Type-annotated monad implementations for Python 3.7+",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/correl/typesafe-monads",
    packages=["monads"],
    package_data={"monads": ["py.typed"]},
    include_package_data=True,
    setup_requires=["pytest-runner"],
    tests_require=["pytest", "mypy", "pytest-cov", "pytest-mypy"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries",
    ],
)



from __future__ import annotations
from typing import Any, Callable, Generic, TypeVar

T = TypeVar("T")
S = TypeVar("S")


class Functor(Generic[T]):
    def map(self, function: Callable[[T], S]) -> Functor[S]:  # pragma: no cover
        raise NotImplementedError

    __mul__ = __rmul__ = map
